import{u as l,W as t,P as r,S as a,n as u}from"./index.3e026b47.js";const s=n=>{const e=l(n.input);return e.loading?t.createElement(r,null):e.result==null?t.createElement(a,null,t.createElement(u,null)):t.createElement("div",{style:{display:"flex",justifyContent:"center",alignItems:"center",height:"100%",width:"100%",textAlign:"center"}},t.createElement("div",{style:{lineHeight:"1em",padding:"0.3em 0.5em",borderRadius:"0.5em",backgroundColor:e.result?"#00A36815":"#FF667016",color:e.result?"#008F5D":"#EB1C45"}},e.result?"True":"False"))};export{s as default};
//# sourceMappingURL=Component.f7c7e1a2.js.map
